import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
from tensorflow.examples.tutorials.mnist import input_data

mnist = input_data.read_data_sets('./MNIST_data', one_hot=True)
sess = tf.InteractiveSession()
x = tf.placeholder(tf.float32, shape=[None, 784])
y_ = tf.placeholder(tf.float32, shape=[None, 10])

# Convolutional layer
x_image = tf.reshape(x, [-1,28,28,1])
W_conv = tf.Variable(tf.truncated_normal([5, 5, 1, 30], stddev=0.1))
b_conv = tf.Variable(tf.constant(0.1, shape=[30]))
h_conv = tf.nn.conv2d(x_image, W_conv, strides=[1, 1, 1, 1], padding='VALID')
h_relu = tf.nn.relu(h_conv + b_conv)
h_pool = tf.nn.max_pool(h_relu, ksize=[1, 2, 2, 1], strides=[1, 2, 2, 1], padding='SAME')

# Fully-connected layer
W_fc1 = tf.Variable(tf.truncated_normal([12 * 12 * 30, 500], stddev=0.1))
b_fc1 = tf.Variable(tf.constant(0.1, shape=[500]))
h_pool_flat = tf.reshape(h_pool, [-1, 12*12*30])
h_fc1 = tf.nn.relu(tf.matmul(h_pool_flat, W_fc1) + b_fc1)

# Output layer
W_fc2 = tf.Variable(tf.truncated_normal([500, 10], stddev=0.1))
b_fc2 = tf.Variable(tf.constant(0.1, shape=[10]))
y_hat=tf.nn.softmax(tf.matmul(h_fc1, W_fc2) + b_fc2)

# Train and Evaluate the Model
cross_entropy = - tf.reduce_sum(y_*tf.log(y_hat))
train_step = tf.train.AdamOptimizer(1e-4).minimize(cross_entropy)
correct_prediction = tf.equal(tf.argmax(y_hat,1), tf.argmax(y_,1))
accuracy = tf.reduce_mean(tf.cast(correct_prediction, tf.float32))

sess.run(tf.initialize_all_variables())
print("=================================")
print("|Epoch\tBatch\t|Train\t|Val\t|")
print("|===============================|")
for j in range(5):
    for i in range(550):
        batch = mnist.train.next_batch(100)
        train_step.run(feed_dict={x: batch[0], y_: batch[1]})
        if i%50 == 49:
            train_accuracy = accuracy.eval(feed_dict={x:batch[0], y_: batch[1]})
            val_accuracy = accuracy.eval(feed_dict=\
                {x: mnist.validation.images, y_:mnist.validation.labels})
            print("|%d\t|%d\t|%.4f\t|%.4f\t|"%(j+1, i+1, train_accuracy, val_accuracy))
print("|===============================|")
test_accuracy = accuracy.eval(feed_dict=\
    {x: mnist.test.images, y_:mnist.test.labels})
print("test accuracy=%.4f"%(test_accuracy))

#==========task3==========
print("Start of Task #3")
psi = np.array(sess.run(h_fc1, feed_dict={x: mnist.test.images}))
pi = psi - np.mean(psi, axis=0)
(W, _, _) = np.linalg.svd(np.matmul(np.transpose(pi), pi))
Z = np.matmul(pi, W)

test_labels = mnist.test.labels
set_size = np.shape(test_labels)[0]
index_arr = np.zeros(100, dtype=int)
i_cnt=0 # this will count index for index_arr[i_cnt]
for i in range(10):
    n_cnt=0 # this will count 10 iterations for each digits
    for j in range(set_size):
        if(n_cnt >= 10):
            break;
        elif(np.argmax(test_labels[j, :]) == i):
            index_arr[i_cnt] = j
            i_cnt+=1
            n_cnt+=1

omega = psi[index_arr, :]
omega = omega - np.mean(omega, axis=0)
Q = np.matmul(omega, W)

#=====plot=====
plt.figure(1)
plt.scatter(Z[:, 0], Z[:, 1], c='darkgray')
plt.scatter(Q[0:10, 0], Q[0:10, 1], c='red')
plt.scatter(Q[10:20, 0], Q[10:20, 1], c='yellowgreen')
plt.scatter(Q[30:40, 0], Q[30:40, 1], c='darkgoldenrod')
plt.scatter(Q[40:50, 0], Q[40:50, 1], c='olivedrab')
plt.scatter(Q[50:60, 0], Q[50:60, 1], c='darkorchid')
plt.scatter(Q[60:70, 0], Q[60:70, 1], c='magenta')
plt.scatter(Q[70:80, 0], Q[70:80, 1], c='aqua')
plt.scatter(Q[80:90, 0], Q[80:90, 1], c='chocolate')
plt.scatter(Q[90:100, 0], Q[90:100, 1], c='indianred')
plt.show()